import sys
from typing import TypeVar

_T = TypeVar('_T', bound=type)


def register_in_module(t: _T) -> _T:
    module = t.__module__
    name = t.__name__
    if hasattr(sys.modules[module], name):
        if getattr(sys.modules[module], name) is not t:
            raise RuntimeError
    else:
        setattr(sys.modules[module], name, t)
    return t



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractNodeLink, AbstractNodeMention, Identifiable
from tdm.abstract.json_schema import generate_model


@generate_model(label='equivalent')
@dataclass(frozen=True)
class EquivalenceNodeLink(Identifiable, AbstractNodeLink[AbstractNodeMention, AbstractNodeMention]):
    """
    Represents a semantic link between two node mentions that refer to the same data (maybe in different forms)
    """
    pass



from dataclasses import dataclass
from typing import Callable, Set, Type, TypeVar

from tdm.abstract.datamodel.identifiable import EnsureIdentifiable


_T = TypeVar('_T', bound='AbstractDomainType')


@dataclass(frozen=True)
class AbstractDomainType(EnsureIdentifiable):
    """
    Base class for knowledge base domain types

    Attributes
    --------
    name:
        The name of the domain type.
    """
    name: str

    def pretty(self, verbose: bool = False) -> str:
        """
        Create a beautiful text representation of object.

        :param verbose: flag to create more detailed representation
        :return: object text representation
        """
        return f"{type(self).__name__}[{self.id}]({self.name})"

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return set()

    @classmethod
    def name_filter(cls: Type[_T], name: str) -> Callable[[_T], bool]:
        def _filter(t: _T) -> bool:
            return t.name == name

        return _filter



from typing import Any, Dict, Optional, Tuple, Union

from pydantic import BaseModel


class MetadataModel(BaseModel):
    def to_metadata(self) -> Dict[str, Any]:
        return self.dict(exclude_none=True)


class FactMetadataModel(MetadataModel):
    created_time: Optional[int]
    modified_time: Optional[int]
    fact_confidence: Optional[Tuple[float]]
    value_confidence: Union[float, Tuple[float, ...], None]  # same as Optional[float, Tuple[float, ...]] (pydantic bug workaround)

    class Config:
        extra = 'allow'  # any other extra fields will be kept


class DocumentMetadataModel(MetadataModel):
    title: Optional[str]
    file_name: Optional[str]
    file_type: Optional[str]
    size: Optional[int]
    created_time: Optional[int]
    access_time: Optional[int]
    modified_time: Optional[int]
    publication_date: Optional[int]
    publication_author: Optional[str]
    description: Optional[str]
    language: Optional[str]
    parent_uuid: Optional[str]
    url: Optional[str]
    platform: Optional[str]
    account: Optional[Tuple[str, ...]]
    access_level: Optional[str]
    user: Optional[str]
    path: Optional[str]
    trust_level: Optional[float]
    markers: Optional[Tuple[str, ...]]
    related_concept_id: Optional[str]
    preview_text: Optional[str]
    story: Optional[str]

    class Config:
        extra = 'allow'  # any other extra fields will be kept



from typing import Type

from tdm.abstract.datamodel.value import AbstractValue
from .abstract import AbstractElementModel, AbstractModelSerializer


class ValueSerializer(AbstractModelSerializer[AbstractValue]):
    def serialize(self, element: AbstractValue) -> AbstractElementModel[AbstractValue]:
        from tdm.json_schema.values import serialize_value
        return serialize_value(element)

    def field_type(self, element_type: Type[AbstractValue]) -> Type[AbstractElementModel[AbstractValue]]:
        from tdm.json_schema.values import VALUE_MODELS
        return VALUE_MODELS[element_type]



from dataclasses import dataclass
from typing import Optional

from tdm.abstract.datamodel.value import AbstractValue, EnsureConfidenced
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class Date(object):
    year: Optional[int] = None
    month: Optional[int] = None
    day: Optional[int] = None


@dataclass(frozen=True)
class Time(object):
    hour: int
    minute: int
    second: int


@dataclass(frozen=True)
class _DateTimeValue(EnsureConfidenced):
    """
    Auxiliary class for `DateTimeValue` to fix dataclass fields order.

    Represents the internal structure of a `DateTimeValue`.

    Attributes
    ----------
    date:
        Normalized date representation.
    time:
        Optional normalized time representation.
    """
    date: Date
    time: Optional[Time] = None


@generate_model
@dataclass(frozen=True)
class DateTimeValue(AbstractValue, _DateTimeValue):
    """
    Represents a normalized KB value for date and time.
    """

    @classmethod
    def from_dict(cls, value: dict) -> 'DateTimeValue':
        """
        Create a `DateTimeValue` object from a dictionary.

        Expected value scheme is as follows:

        {
          "date": {
            "year": Optional[int],
            "month": Optional[int],
            "day": Optional[int]
          },
          "time": {
            "hour": int,
            "minute": int,
            "second": int
          }
        }

        Time part could be skipped.

        :param value: The dictionary containing the DateTimeValue information.
        :return: A DateTimeValue object.
        """
        args = {
            'date': Date(**value['date'])
        }
        if 'time' in value:
            args['time'] = Time(**value['time'])
        return cls(**args)



from dataclasses import dataclass

from tdm.abstract.datamodel.value import AbstractValue, EnsureConfidenced
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class _LinkValue(EnsureConfidenced):
    """
    Auxiliary class for `LinkValue` to fix dataclass fields order.

    Represents the internal structure of a `LinkValue`.

    Attributes
    ----------
    link:
        Normalized URL.
    """
    link: str


@generate_model
@dataclass(frozen=True)
class LinkValue(AbstractValue, _LinkValue):
    """
    Represents a normalized KB value for URL.
    """
    pass



__all__ = [
    'cache_result',
    'freeze', 'freeze_dict', 'freeze_sequence', 'unfreeze',
    'register_in_module',
    'check_base_type', 'generics_mapping', 'is_subclass', 'unfold_union', 'uniform_collection'
]

from .cache import cache_result
from .immutability import freeze, freeze_dict, freeze_sequence, unfreeze
from .module import register_in_module
from .typing import check_base_type, generics_mapping, is_subclass, unfold_union, uniform_collection



from typing import Collection, Dict, Generic, Mapping, Optional, Tuple, TypeVar, Union


def unfold_union(type_: type) -> Tuple[type, ...]:
    # check if it is union
    if hasattr(type_, '__origin__') and type_.__origin__ is Union:
        return type_.__args__
    return type_,


def uniform_collection(t) -> Tuple[Optional[type], type]:
    if not hasattr(t, '__origin__'):  # not a generic, just plain type
        return None, t
    if issubclass(t.__origin__, Mapping):
        key_type, value_type = t.__args__
        if key_type is not str:
            raise TypeError(f"Only Mapping[str, ...] coul be serialized. Actual: {t}")
        return t.__origin__, value_type
    if issubclass(t.__origin__, Collection):
        args = tuple(filter(lambda t: t is not ..., t.__args__))
        if len(args) != 1:
            raise NotImplementedError
        return t.__origin__, args[0]
    if t.__origin__ is type:  # Type[...]
        return type, t.__args__[0]
    raise TypeError


def check_base_type(type_: type, base_type: type) -> bool:
    if issubclass(type_, base_type):  # check simple subclass
        return True
    if not hasattr(type_, '__origin__'):  # check if it is some generic (Union, Tuple, etc)
        return False
    origin = type_.__origin__
    if not issubclass(origin, Tuple):  # check if it is tuple
        return False
    for arg in filter(lambda t: t is not ..., type_.__args__):  # check each argument except Ellipsis
        if not issubclass(arg, base_type):  # Now we assume only Tuple[type_, ...]
            return False
    return True


def is_subclass(type_: type, base_type: type) -> bool:
    try:
        return all(check_base_type(t, base_type) for t in unfold_union(type_))
    except TypeError:
        return False


def generics_mapping(type_: type) -> Dict[TypeVar, type]:
    if issubclass(type_, Generic):  # actually it works (should be rewritten in more pythonic way)
        # assume type vars are not intersect
        result = {}
        for orig_base in type_.__orig_bases__:
            if '__origin__' in orig_base.__dict__ and '__args__' in orig_base.__dict__:
                if not hasattr(orig_base.__origin__, '__parameters__'):
                    print('here')
                result.update(dict(zip(orig_base.__origin__.__parameters__, orig_base.__args__)))
        return result
    return {}



from abc import ABCMeta, abstractmethod
from typing import Callable, Dict, Iterable, Iterator, Optional, Set, Tuple, Type, TypeVar, Union

from typing_extensions import Self

from .fact import AbstractFact
from .identifiable import Identifiable
from .node import AbstractNode
from .node_link import AbstractNodeLink

_Fact = TypeVar('_Fact', bound=AbstractFact)
_NodeLink = TypeVar('_NodeLink', bound=AbstractNodeLink)
_Node = TypeVar('_Node', bound=AbstractNode)

NodeOrId = Union[str, AbstractNode]


class TalismanDocument(metaclass=ABCMeta):
    __slots__ = ()

    @property
    @abstractmethod
    def id(self) -> str:
        """
        Document identifier.
        """
        pass

    # nodes methods

    @property
    @abstractmethod
    def id2node(self) -> Dict[str, AbstractNode]:
        """
        Get all document nodes.

        :return: mapping from node identifier to node
        """
        pass

    @property
    @abstractmethod
    def nodes(self) -> Dict[Type[AbstractNode], Iterable[AbstractNode]]:
        """
        Get all document nodes.

        :return: mapping from node type to nodes of corresponding type
        """
        pass

    @abstractmethod
    def get_node(self, id_: str) -> AbstractNode:
        """
        Get node by its identifier.
        If no such node contained in the document, ``KeyError`` will be raised.

        :param id_: node identifier
        :return: node
        """
        pass

    @abstractmethod
    def get_nodes(
            self, type_: Type[_Node] = AbstractNode, *,
            filter_: Union[Callable[[_Node], bool], Iterable[Callable[[_Node], bool]]] = tuple()
    ) -> Iterator[_Node]:
        """
        Get all nodes of specified type that satisfy the filter condition.

        :param type_: type of the desired nodes
        :param filter_: predicate to test each individual document node
        :return: iterator of the nodes
        """
        pass

    @abstractmethod
    def related_nodes(
            self, obj: Union[Identifiable, str], type_: Type[_Node] = AbstractNode, *,
            filter_: Union[Callable[[_Node], bool], Iterable[Callable[[_Node], bool]]] = tuple()
    ) -> Iterator[_Node]:
        """
        Get all nodes of specified type that depend on identifiable object (nodes, facts, etc.) and satisfy the filter condition.

        :param obj: identifier or identifiable object (only identifier matters) to filter nodes.
        :param type_: type of the desired nodes
        :param filter_: predicate to test each individual node
        :return: iterator of the nodes
        """
        pass

    @abstractmethod
    def with_nodes(self, nodes: Iterable[AbstractNode]) -> Self:
        """
        Create the same document with new nodes.
        This method replaces nodes already contained in the document if it is possible. Otherwise, ``ValueError`` is risen.
        Each node type declare a set of fields that couldn't be changed (see ``AbstractNode.constant_fields`` method).

        :param nodes: document nodes to be added/updated
        :return: document with nodes added
        """
        pass

    @abstractmethod
    def without_nodes(self, nodes: Iterable[NodeOrId], *, cascade: bool = True) -> Self:
        """
        Create the same document without specified nodes.
        Only identifiers are taken into account.
        In other words, if document contains ``ImageNode(..., id='123')``, and ``TextNode(..., id='123')`` is tried to be removed,
        ``ImageNode`` will be successfully removed without any warnings.
        All nodes not contained in the document are skipped silently.

        Enabling ``cascade`` flag leads to subsequently remove all document objects that depends on nodes to be removed
        (e.g. semantic links or mention facts).
        If no ``cascade`` flag is enabled, and some object depends on node to be removed, ``ValueError`` will be risen.

        :param nodes: iterable of `AbstractNode`s or node identifiers
        :param cascade: enable cascade removing (all objects associated with specified nodes will be removed too)
        :return: the same document without specified nodes and associated objects
        """
        pass

    # structure methods

    @property
    @abstractmethod
    def roots(self) -> Set[AbstractNode]:
        """
        Get set of nodes that have not parent nodes.
        This property is derived from document structure automatically.
        """
        pass

    @property
    @abstractmethod
    def main_root(self) -> Optional[AbstractNode]:
        """
        Get main root node.
        Main node should be set manually (via ``TalismanDocument.with_main_root`` method).
        """
        pass

    @abstractmethod
    def with_main_root(self, node: Optional[NodeOrId], *, update: bool = False) -> Self:
        """
        Create the same document with node set to be main root.
        Document without main root can't be properly serialized.
        Enabling ``update`` flag is the same as subsequent calls ``with_nodes([node]).with_main_root(node.id)``.

        :param node: node or node identifier to be set as main root
        :param update: flag to update node in the document (not valid for node identifier)
        :return: the same document with main root and optionally updated node
        """
        pass

    @abstractmethod
    def parent(self, node: NodeOrId) -> Optional[AbstractNode]:
        """
        Get parent node of the specified node (if exists)

        :param node: node or node id
        :return: direct parent node or None
        """
        pass

    @abstractmethod
    def child_nodes(self, node: Union[str, AbstractNode]) -> Tuple[AbstractNode, ...]:
        """
        Get children nodes for specified node

        :param node: node or node id
        :return: tuple of direct children nodes
        """
        pass

    @abstractmethod
    def with_structure(self, structure: Dict[NodeOrId, Iterable[NodeOrId]], *, force: bool = False, update: bool = False) -> Self:
        """
        Create the same document with updated structure.
        Structure is a mapping from node to iterable of children nodes.
        Each node could have the only parent node. If some node is marked as main root, it couldn't have any parent.
        Circular dependencies are not permitted.

        Enabling ``force`` flag leads to all conflict dependencies overwriting (except circular dependencies that couldn't be resolved
        automatically).
        If ``force`` is enabled, all conflict child->parent links presented in original document will be removed in result document.
        In any case structure mapping should not be self-conflicting.

        Enabling ``update`` flag leads to nodes replacement (or adding).

        :param structure: mapping from parent node (or identifier) to children nodes (or its identifiers)
        :param force: enable structure overwriting
        :param update: enable specified nodes updates
        :return: the same document with structure updated
        """
        pass

    @abstractmethod
    def with_node_parent(self, child: NodeOrId, parent: NodeOrId, *, force: bool = False, update: bool = False) -> Self:
        """
        Create the same document with updated structure.
        Child node is appended as last children of the parent node.
        Each node could have the only parent node.
        If some node is marked as main root, it couldn't have any parent.
        Circular dependencies are not permitted.

        Enabling ``force`` flag leads to child's parent overwriting.

        Enabling ``update`` flag leads to both parent and child nodes replacement (or adding).

        :param child: child node or node identifier
        :param parent: parent node or node identifier
        :param force: enable structure overwriting
        :param update: enable child and parent nodes updates
        :return: the same document with structure updated
        """
        pass

    @abstractmethod
    def with_roots(self, nodes: Iterable[NodeOrId], *, update: bool = False) -> Self:
        """
        Create the same document with updated structure.
        Removes parent links for specified nodes.

        :param nodes: nodes or node identifiers to be set as roots
        :param update: enable passed nodes to be updated
        :return: the same document with structure updated
        """
        pass

    # semantic link methods

    @property
    @abstractmethod
    def id2node_link(self) -> Dict[str, AbstractNodeLink]:
        """
        Get all document node semantic links

        :return: mapping from link identifier to link
        """
        pass

    @property
    @abstractmethod
    def node_links(self) -> Dict[Type[AbstractNodeLink], Iterable[AbstractNodeLink]]:
        """
        Get all document node semantic links

        :return: mapping from link type to iterable of corresponding type links
        """
        pass

    @abstractmethod
    def get_node_link(self, id_: str) -> AbstractNodeLink:
        """
        Get node link by its identifier.
        If no such link contained in the document, ``KeyError`` will be raised.

        :param id_: link identifier
        :return: semantic link
        """
        pass

    @abstractmethod
    def get_node_links(
            self, type_: Type[_NodeLink] = AbstractNodeLink, *,
            filter_: Union[Callable[[_NodeLink], bool], Iterable[Callable[[_NodeLink], bool]]] = tuple()
    ) -> Iterator[_NodeLink]:
        """
        Get all semantic links of specified type that satisfy the filter condition.

        :param type_: type of the desired links
        :param filter_: predicate to test each individual semantic link
        :return: iterator of the links
        """
        pass

    @abstractmethod
    def related_node_links(
            self, obj: Union[Identifiable, str], type_: Type[_NodeLink] = AbstractNodeLink, *,
            filter_: Union[Callable[[_NodeLink], bool], Iterable[Callable[[_NodeLink], bool]]] = tuple()
    ) -> Iterator[_NodeLink]:
        """
        Get all semantic links of specified type that depend on identifiable object (nodes, facts, etc.) and satisfy the filter condition.

        :param obj: identifier or identifiable object (only identifier matters) to filter semantic links
        :param type_: type of the desired links
        :param filter_: predicate to test each individual semantic link
        :return: iterator of the links
        """
        pass

    @abstractmethod
    def with_node_links(self, links: Iterable[AbstractNodeLink], *, update: bool = False) -> Self:
        """
        Create the same document with new node semantic links.
        This method replaces links already contained in the document if it is possible. Otherwise, ``ValueError`` is risen.
        Each semantic link type declare a set of fields that couldn't be changed (see AbstractNodeLink.constant_fields method).

        Enabling ``update`` flag leads to adding or updating all the objects, stored in specified links (e.g. nodes).

        :param links: document node semantic links to be added/updated
        :param update: add or update all identifiable objects contained in specified node semantic links
        :return: document with node semantic links added
        """
        pass

    @abstractmethod
    def without_node_links(self, links: Iterable[Union[str, AbstractNodeLink]], *, cascade: bool = False) -> Self:
        """
        Create the same document without specified node semantic links.
        Only identifiers are taken into account.
        In other words, if document contains ``EquivalenceNodeLink(..., id='123')``, and ``TranslationNodeLink(..., id='123')``
        is tried to be removed, ``EquivalenceNodeLink`` will be successfully removed without any warnings.
        All links not contained in the document are skipped silently.

        Enabling ``cascade`` flag leads to subsequently remove all document objects that depends on links to be removed.
        If no ``cascade`` flag is enabled, and some object depends on link to be removed, ``ValueError`` will be risen.

        :param links: iterable of ``AbstractNodeLink`` or link identifiers
        :param cascade: enable cascade removing (all objects associated with specified links will be removed too)
        :return: the same document without specified links and associated objects
        """
        pass

    # facts methods

    @property
    @abstractmethod
    def id2fact(self) -> Dict[str, AbstractFact]:
        """
        Get all document facts

        :return: mapping from fact identifier to fact
        """
        pass

    @property
    @abstractmethod
    def facts(self) -> Dict[Type[AbstractFact], Iterable[AbstractFact]]:
        """
        Get all document facts

        :return: mapping from fact type to iterable of corresponding type facts
        """
        pass

    @abstractmethod
    def get_fact(self, id_: str) -> AbstractFact:
        """
        Get fact by its identifier. If no such fact contained in the document, ``KeyError`` will be raised.

        :param id_: fact identifier
        :return: fact
        """
        pass

    @abstractmethod
    def get_facts(
            self, type_: Type[_Fact] = AbstractFact, *,
            filter_: Union[Callable[[_Fact], bool], Iterable[Callable[[_Fact], bool]]] = tuple()
    ) -> Iterator[_Fact]:
        """
        Get all facts of specified type that satisfy the filter condition.

        :param type_: type of the desired facts
        :param filter_: predicate to test each individual fact
        :return: iterator of the facts
        """
        pass

    @abstractmethod
    def related_facts(
            self, obj: Union[Identifiable, str], type_: Type[_Fact] = AbstractFact, *,
            filter_: Union[Callable[[_Fact], bool], Iterable[Callable[[_Fact], bool]]] = tuple()
    ) -> Iterator[_Fact]:
        """
        Get all facts of specified type that depend on identifiable object (nodes, facts, etc.) and satisfy the filter condition.

        :param obj: identifier or identifiable object (only identifier matters) to filter facts
        :param type_: type of the desired facts
        :param filter_: predicate to test each individual fact
        :return: iterator of the facts
        """
        pass

    @abstractmethod
    def with_facts(self, facts: Iterable[AbstractFact], *, update: bool = False) -> Self:
        """
        Create the same document with new facts.
        This method replaces facts already contained in the document if it is possible. Otherwise, ``ValueError`` is risen.
        Each fact type declare a set of fields that couldn't be changed (see ``AbstractFact.constant_fields method``).

        Enabling ``update`` flag leads to adding or updating all the objects, stored in specified facts (e.g. other facts, nodes).

        :param facts: document facts to be added/updated
        :param update: add or update all identifiable objects contained in specified facts
        :return: document with facts added
        """
        pass

    @abstractmethod
    def without_facts(self, facts: Iterable[Union[str, AbstractFact]], *, cascade: bool = False) -> Self:
        """
        Create the same document without specified facts.
        Only identifiers are taken into account.
        In other words, if document contains ``ConceptFact(..., id='123')``, and ``PropertyFact(..., id='123')`` is tried to be removed,
        ``ConceptFact`` will be successfully removed without any warnings.
        All links not contained in the document are skipped silently.

        Enabling ``cascade`` flag leads to subsequently remove all document objects that depends on facts to be removed.
        If no ``cascade`` flag is enabled, and some object depends on fact to be removed, ``ValueError`` will be risen.

        :param facts: iterable of ``AbstractFact`` or fact identifiers
        :param cascade: enable cascade removing (all objects associated with specified facts will be removed too)
        :return: the same document without specified facts and associated objects
        """
        pass


class AbstractDocumentFactory(metaclass=ABCMeta):
    @abstractmethod
    def create_document(self, *, id_: Optional[str] = None, doc_type: Optional[str] = None) -> TalismanDocument:
        """
        create empty document with specified identifier.
        If no identifier specified, it will be generated automatically.

        If doc_type is present, document fact will be added to the document.

        :param id_: new document identifier
        :param doc_type: document type identifier
        :return: empty document
        """
        pass



from dataclasses import dataclass, field
from typing import Generic, Set, TypeVar

from tdm.helper import generics_mapping
from .identifiable import Identifiable
from .markup import AbstractMarkup, FrozenMarkup


@dataclass(frozen=True)
class BaseNodeMetadata:
    """
    The most base node metadata

    Attributes
    ---------
    hidden:
        node hiding indicator
    """
    hidden: bool = False


_NodeMetadata = TypeVar("_NodeMetadata", bound=BaseNodeMetadata)


@dataclass(frozen=True)
class AbstractNode(Identifiable, Generic[_NodeMetadata]):
    """
    Base interface for document nodes.
    All the document nodes should be frozen dataclasses.

    Attributes
    --------
    metadata:
        node metadata (defined by node implementation dataclass with fixed field set)
    markup:
        any structured node markup (no fixed schema)
    """

    metadata: _NodeMetadata = None
    markup: AbstractMarkup = field(default_factory=FrozenMarkup, hash=False)

    def __post_init__(self):
        super().__post_init__()
        if self.metadata is None:
            # hack for runtime metadata generation (if no value passed)
            object.__setattr__(self, 'metadata', self._generate_metadata())
        object.__setattr__(self, 'markup', self._convert_markup(self.markup))

    def _convert_markup(self, markup: AbstractMarkup) -> AbstractMarkup:
        if not isinstance(markup, AbstractMarkup):
            raise ValueError
        return markup

    def _generate_metadata(self) -> _NodeMetadata:
        type_vars = generics_mapping(type(self))
        return type_vars[_NodeMetadata]()

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return set()


_Content = TypeVar('_Content')


@dataclass(frozen=True)
class _AbstractContentNode(Generic[_Content]):
    """
    Auxiliary class for ``AbstractContentNode`` to fix dataclass fields order

    Attributes
    --------
    content:
        Node content
    """
    content: _Content


@dataclass(frozen=True)
class AbstractContentNode(AbstractNode[_NodeMetadata], _AbstractContentNode[_Content], Generic[_NodeMetadata, _Content]):
    """
    Base class for all data nodes.
    """

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'content'}



from dataclasses import dataclass
from typing import Type

from tdm.abstract.datamodel import AbstractDomainType, Identifiable
from tdm.abstract.datamodel.domain import AbstractValueDomainType
from tdm.abstract.datamodel.value import AbstractConceptValue


@dataclass(frozen=True)
class AbstractConceptType(Identifiable, AbstractValueDomainType[AbstractConceptValue]):
    """
    Base class for concept-like domain types
    """
    value_type: Type[AbstractConceptValue] = AbstractConceptValue

    def pretty(self, verbose: bool = False) -> str:
        return AbstractDomainType.pretty(self, verbose)


@dataclass(frozen=True)
class ConceptType(AbstractConceptType):
    """
    Concept domain type
    """
    pass


@dataclass(frozen=True)
class DocumentType(AbstractConceptType):
    """
    Concept domain type for document representation
    """
    pass


#  following classes could be removed after v0 support stop

@dataclass(frozen=True)
class AccountType(AbstractConceptType):
    """
    Concept domain type for account representation
    """
    pass


@dataclass(frozen=True)
class PlatformType(AbstractConceptType):
    """
    Concept domain type for platform representation
    """
    pass



from queue import LifoQueue
from typing import Callable, Iterable, Iterator, Optional, Type, TypeVar, Union

from typing_extensions import Protocol

from tdm import TalismanDocument, and_filter
from tdm.abstract.datamodel import AbstractNode

_Node = TypeVar('_Node', bound=AbstractNode)


class StopCriteria(Protocol):
    def __call__(self, node: AbstractNode) -> bool:
        pass


class NodeFilter(Protocol):
    def __call__(self, node: _Node) -> bool:
        pass


def _dfs(document: TalismanDocument, root: AbstractNode, stop_criteria: Optional[StopCriteria] = None) -> Iterator[AbstractNode]:
    if stop_criteria is None:
        def stop_criteria(_: AbstractNode) -> bool:
            return False

    queue = LifoQueue()
    queue.put(root)

    while not queue.empty():
        dfs_lookup_node = queue.get()
        yield dfs_lookup_node
        if stop_criteria(dfs_lookup_node):
            continue

        # new nodes are visited left to right to preserve reading order
        for child_node in reversed(document.child_nodes(dfs_lookup_node)):
            queue.put(child_node)


def dfs(
        document: TalismanDocument,
        root: Optional[Union[AbstractNode, Iterable[AbstractNode]]] = None,
        type_: Type[_Node] = AbstractNode,
        *,
        stop_criteria: Optional[StopCriteria] = None,
        filter_: Union[NodeFilter, Iterable[NodeFilter]] = tuple()
) -> Iterator[_Node]:
    """
    Perform a Depth-First Search (DFS) on the nodes of a Talisman document.

    :param document: The document to traverse.
    :param root: The starting node or nodes for the DFS. If no starting node is specified, document main root is used as starting node.
    :param type_: The type of nodes to consider during the search.
    :param stop_criteria: A criteria function to determine when to stop the search.
    :param filter_: Node filters to apply during the search.
    :return: An iterator over the nodes of specified type that satisfy the filter.
    """

    def check_type(obj):
        return isinstance(obj, type_)

    if isinstance(filter_, Callable):
        filter_ = (filter_,)
    filter_ = and_filter(check_type, *filter_)

    if root is None:
        root = document.main_root

    if not isinstance(root, Iterable):
        root = (root,)

    for start_node in root:
        yield from filter(filter_, _dfs(document, start_node, stop_criteria))



from typing import Callable, Dict, Iterable, Iterator, Type, TypeVar, Union

from tdm.abstract.datamodel import AbstractDomain, AbstractDomainType, Identifiable
from tdm.datamodel.common import TypedIdsContainer, ViewContainer

_DomainType = TypeVar('_DomainType', bound=AbstractDomainType)


class Domain(AbstractDomain, ViewContainer):
    def __init__(
            self,
            types: Iterable[AbstractDomainType]
    ):
        temp = ViewContainer({}, {}, {AbstractDomainType: TypedIdsContainer(AbstractDomainType, ())}).with_elements(types, update=True)
        super().__init__(temp._id2view, temp._dependencies, temp._containers)

    @property
    def id2type(self) -> Dict[str, AbstractDomainType]:
        return self.id2element(AbstractDomainType)

    @property
    def types(self) -> Dict[Type[AbstractDomainType], Iterable[AbstractDomainType]]:
        return self.elements(AbstractDomainType)

    def get_type(self, id_: str) -> AbstractDomainType:
        return self.get_element(AbstractDomainType, id_)

    def get_types(
            self, type_: Type[_DomainType] = AbstractDomainType, *,
            filter_: Union[Callable[[_DomainType], bool], Iterable[Callable[[_DomainType], bool]]] = tuple()
    ) -> Iterator[_DomainType]:
        return self.get_elements(AbstractDomainType, type_, filter_=filter_)

    def related_types(
            self, obj: Union[Identifiable, str], type_: Type[_DomainType] = AbstractDomainType, *,
            filter_: Union[Callable[[_DomainType], bool], Iterable[Callable[[_DomainType], bool]]] = tuple()
    ) -> Iterator[_DomainType]:
        return self.related_elements(AbstractDomainType, obj, type_, filter_=filter_)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Domain):
            return NotImplemented
        return self._id2view == o._id2view and self._containers == o._containers

    def __hash__(self) -> int:
        return hash(self._containers[AbstractDomainType])



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractNodeMention
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.nodes import TextNode


@generate_model
@dataclass(frozen=True)
class TextNodeMention(AbstractNodeMention):
    """
    Represents text node continuous mention [start; end)

    Attributes
    --------
    node:
        mentioned text node
    start:
        index of start symbol (inclusive)
    end:
        index of end symbol (exclusive)
    """
    node: TextNode
    start: int
    end: int

    def __post_init__(self):
        if not isinstance(self.node, TextNode):
            raise ValueError(f"Incorrect node type {type(self.node)}. Expected {TextNode}")
        if self.start < 0 or self.end <= self.start:
            raise ValueError(f"Incorrect span [{self.start}, {self.end})")
        if self.end > len(self.node.content):
            raise ValueError(f"Span spreads out of the text (text len: {len(self.node.content)})")



from abc import ABCMeta
from dataclasses import dataclass, replace
from typing import Callable, Set, Union

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, Identifiable
from tdm.abstract.datamodel.fact import AbstractValueFact
from tdm.abstract.datamodel.value import AbstractValue
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import AtomValueType, CompositeValueType


@generate_model(label='atom')
@dataclass(frozen=True)
class AtomValueFact(Identifiable, AbstractValueFact[AtomValueType, AbstractValue]):
    """
    Represents a fact that holds an atomic value.

    If the fact status is FactStatus.NEW `value` should be a tuple.
    If the fact status is FactStatus.APPROVED `value` should be single-valued.
    """
    def replace_with_domain(self, domain: AbstractDomain) -> 'AtomValueFact':
        if isinstance(self.type_id, str):
            domain_type = domain.get_type(self.type_id)
            if not isinstance(domain_type, AtomValueType):
                raise ValueError
            return replace(self, type_id=domain_type)
        return self

    def _as_tuple(self) -> tuple:
        return self.id, (self.type_id if isinstance(self.type_id, str) else self.type_id.id), self.value

    def __eq__(self, other):
        if not isinstance(other, AtomValueFact):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    @staticmethod
    def empty_value_filter() -> Callable[['AtomValueFact'], bool]:
        """
        Create a filter function to filter `AtomValueFact` instances with empty values.

        :return: A filter function for `AtomValueFact` instances with empty values.
        """
        return lambda f: isinstance(f.value, tuple) and not f.value

    @staticmethod
    def tuple_value_filter() -> Callable[['AtomValueFact'], bool]:
        """
        Create a filter function to filter `AtomValueFact` instances with tuple values.

        :return: A filter function for `AtomValueFact` instances with tuple values.
        """
        return lambda f: isinstance(f.value, tuple)

    @staticmethod
    def single_value_filter() -> Callable[['AtomValueFact'], bool]:
        """
        Create a filter function to filter `AtomValueFact` instances with a single value.

        :return: A filter function for `AtomValueFact` instances with a single value.
        """
        return lambda f: isinstance(f.value, AbstractValue)


@dataclass(frozen=True)
class _CompositeValueFact(AbstractFact, metaclass=ABCMeta):
    """
    Auxiliary class for `CompositeValueFact` to fix dataclass fields order.

    Attributes
    ----------
    type_id:
        The type identifier or domain composite value type of the composite value fact.
    """
    type_id: Union[str, CompositeValueType]

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}

    @property
    def str_type_id(self) -> str:
        return self.type_id if isinstance(self.type_id, str) else self.type_id.id


@generate_model(label='composite')
@dataclass(frozen=True)
class CompositeValueFact(Identifiable, _CompositeValueFact):
    """
    Represents a fact that holds a composite value.

    The whole value is represented as `CompositeValueFact` with related `ComponentFact`s.
    """

    def replace_with_domain(self, domain: AbstractDomain) -> 'CompositeValueFact':
        if isinstance(self.type_id, str):
            domain_type = domain.get_type(self.type_id)
            if not isinstance(domain_type, CompositeValueType):
                raise ValueError
            return replace(self, type_id=domain_type)
        return self

    def _as_tuple(self) -> tuple:
        return self.id, (self.type_id if isinstance(self.type_id, str) else self.type_id.id)

    def __eq__(self, other):
        if not isinstance(other, CompositeValueFact):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())


ValueFact = Union[AtomValueFact, CompositeValueFact]



from abc import ABCMeta
from dataclasses import dataclass, replace
from typing import Callable, Set, Union

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, AbstractNode, AbstractNodeMention, Identifiable
from tdm.abstract.json_schema import generate_model
from .value import AtomValueFact


@dataclass(frozen=True)
class _MentionFact(AbstractFact, metaclass=ABCMeta):
    """
    Auxiliary class for `MentionFact` to fix dataclass fields order.

    Attributes
    ----------
    mention:
        The part of the node that contains value fact mention.
    value:
        The value fact mentioned in document node.
    """
    mention: AbstractNodeMention
    value: AtomValueFact

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'mention', 'value'}


@generate_model(label='mention')
@dataclass(frozen=True)
class MentionFact(Identifiable, _MentionFact):
    """
    Represents fact that some `AtomValueFact` is mentioned in the document node.

    """

    def replace_with_domain(self, domain: AbstractDomain) -> 'MentionFact':
        value = self.value.replace_with_domain(domain)
        if value is self.value:
            return self
        return replace(self, value=value)

    @staticmethod
    def node_filter(node: Union[AbstractNode, str]) -> Callable[['MentionFact'], bool]:
        """
        Create a filter function to filter `MentionFact` instances based on the document node.

        :param node: The document node or its identifier to filter by.
        :return: A filter function for `MentionFact` instances.
        """
        node_id = node.id if isinstance(node, AbstractNode) else node

        def _filter(fact: MentionFact) -> bool:
            return fact.mention.node_id == node_id

        return _filter

    @staticmethod
    def value_filter(filter_: Callable[[AtomValueFact], bool]) -> Callable[['MentionFact'], bool]:
        """
        Create a filter function to filter `MentionFact` instances based on the value.

        :param filter_: A filter function for mentioned `AtomValueFact`.
        :return: A filter function for `MentionFact` instances.
        """

        def _filter(fact: MentionFact) -> bool:
            return filter_(fact.value)

        return _filter



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractLinkFact, Identifiable
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import ComponentValueType, RelationPropertyType, RelationType
from .concept import ConceptFact
from .value import CompositeValueFact, ValueFact


@generate_model(label='relation')
@dataclass(frozen=True, eq=False)
class RelationFact(Identifiable, AbstractLinkFact[ConceptFact, ConceptFact, RelationType]):
    """
    Represents a relation fact that links together two concept facts.
    """
    pass


@generate_model(label='r_property')
@dataclass(frozen=True, eq=False)
class RelationPropertyFact(Identifiable, AbstractLinkFact[RelationFact, ValueFact, RelationPropertyType]):
    """
    Represents a relation property fact.
    It links relation fact with some value fact.
    """
    pass


@generate_model(label='component')
@dataclass(frozen=True, eq=False)
class ComponentFact(Identifiable, AbstractLinkFact[CompositeValueFact, ValueFact, ComponentValueType]):
    """
    Represents a composite value component fact.
    It links composite value fact with another value fact (its component).
    """
    pass



import inspect
from functools import wraps
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Set, Tuple, Type, TypeVar, Union

from typing_extensions import ParamSpec, Self

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, AbstractNode, AbstractNodeLink, EnsureIdentifiable, \
    Identifiable, TalismanDocument
from tdm.datamodel.common import AbstractView, TypedIdsContainer, ViewContainer, restore_object
from ._structure import NodesStructure

_D = TypeVar('_D', bound=EnsureIdentifiable)

_Node = TypeVar('_Node', bound=AbstractNode)
_NodeLink = TypeVar('_NodeLink', bound=AbstractNodeLink)
_Fact = TypeVar('_Fact', bound=AbstractFact)

_I = TypeVar('_I', bound=Identifiable)

_P = ParamSpec('_P')


def _filter_elements(base_type: Type[_D]):
    def decorator(f: Callable[_P, 'TalismanDocumentImpl']) -> Callable[_P, 'TalismanDocumentImpl']:
        signature = inspect.signature(f)
        target_parameter = tuple(signature.parameters)[1]

        @wraps(f)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> 'TalismanDocumentImpl':
            arguments = signature.bind(*args, **kwargs).arguments
            self = arguments['self']
            arguments[target_parameter] = _filter(self._containers[base_type], self._id2view, arguments[target_parameter], base_type)
            return f(**arguments)

        return wrapper

    return decorator


class TalismanDocumentImpl(TalismanDocument, ViewContainer):
    __slots__ = (
        '_id', '_domain', '_structure'
    )

    def __init__(
            self,
            id2view: Dict[str, Union[EnsureIdentifiable, AbstractView]],
            dependencies: Dict[str, Set[Tuple[str, Type[EnsureIdentifiable]]]],
            containers: Dict[Type[EnsureIdentifiable], TypedIdsContainer],
            structure: NodesStructure,
            *, id_: str, domain: Optional[AbstractDomain] = None):
        super().__init__(id2view, dependencies, containers)
        self._id = id_
        self._domain = domain
        self._structure = structure

    def _replace(self, **kwargs) -> Self:
        return super()._replace(
            **{
                'structure': self._structure,
                'id_': self._id,
                'domain': self._domain,
                **kwargs
            }
        )

    def _update_structure(self, other: ViewContainer) -> Self:
        if not isinstance(other, TalismanDocumentImpl):
            raise ValueError
        if other._containers[AbstractNode] is not self._containers[AbstractNode]:  # something has changed
            other._structure = self._structure.update_nodes(other._containers[AbstractNode].ids)
        return other

    def with_elements(self, elements: Iterable[EnsureIdentifiable], *, update: bool = False) -> Self:
        return self._update_structure(super().with_elements(elements, update=update))

    def without_elements(self, ids: Iterable[str], *, cascade: bool = False) -> Self:
        return self._update_structure(super().without_elements(ids, cascade=cascade))

    def _transform_element(self, element: _D) -> _D:
        if self._domain is not None and isinstance(element, AbstractFact):
            return element.replace_with_domain(self._domain)
        return element

    @property
    def id(self) -> str:
        return self._id

    @property
    def id2node(self) -> Dict[str, AbstractNode]:
        return self.id2element(AbstractNode)

    @property
    def nodes(self) -> Dict[Type[AbstractNode], Iterable[AbstractNode]]:
        return self.elements(AbstractNode)

    def get_node(self, id_: str) -> AbstractNode:
        return self.get_element(AbstractNode, id_)

    def get_nodes(
            self, type_: Type[_Node] = AbstractNode, *,
            filter_: Union[Callable[[_Node], bool], Iterable[Callable[[_Node], bool]]] = tuple()
    ) -> Iterator[_Node]:
        return self.get_elements(AbstractNode, type_, filter_=filter_)

    def related_nodes(
            self, obj: Union[Identifiable, str], type_: Type[_Node] = AbstractNode, *,
            filter_: Union[Callable[[_Node], bool], Iterable[Callable[[_Node], bool]]] = tuple()
    ) -> Iterator[_Node]:
        return self.related_elements(AbstractNode, obj, type_, filter_=filter_)

    @_filter_elements(AbstractNode)
    def with_nodes(self, nodes: Iterable[AbstractNode]) -> Self:
        return self.with_elements(nodes, update=True)

    @_filter_elements(AbstractNode)
    def without_nodes(self, nodes: Iterable[Union[str, AbstractNode]], *, cascade: bool = False) -> Self:
        return self.without_elements(nodes, cascade=cascade)

    # structure methods

    @property
    def roots(self) -> Set[AbstractNode]:
        return {restore_object(self._id2view[id_], self._id2view) for id_ in self._structure.roots}

    @property
    def main_root(self) -> Optional[AbstractNode]:
        main_root = self._structure.main_root
        return restore_object(self._id2view[main_root], self._id2view) if main_root is not None else None

    def with_main_root(self, node: Optional[Union[str, AbstractNode]], *, update: bool = False) -> Self:
        if node is None:
            return self._replace(structure=self._structure.with_main_root(None))
        if update:
            if not isinstance(node, AbstractNode):
                raise ValueError(f"could not update node {node} with id only")
            result = self.with_nodes((node,))
            result._structure = result._structure.with_main_root(node.id)
            return result
        return self._replace(structure=self._structure.with_main_root(_to_id(node)))

    def parent(self, node: Union[str, AbstractNode]) -> Optional[AbstractNode]:
        node_id = _to_id(node)
        parent_id = self._structure.parent.get(node_id)
        if parent_id is not None:
            return restore_object(self._id2view[parent_id], self._id2view)
        return None

    def child_nodes(self, node: Union[str, AbstractNode]) -> Tuple[AbstractNode, ...]:
        return tuple(restore_object(self._id2view[id_], self._id2view) for id_ in self._structure.children.get(_to_id(node), ()))

    def with_structure(
            self, structure: Dict[Union[str, AbstractNode], Iterable[Union[str, AbstractNode]]],
            *, force: bool = False, update: bool = False
    ) -> Self:
        _structure = {}
        nodes = set()

        for parent, children in structure.items():
            children = list(children)
            if update:
                if isinstance(parent, AbstractNode):
                    nodes.add(parent)
                nodes.update(node for node in children if isinstance(node, AbstractNode))
            _structure[_to_id(parent)] = [_to_id(node) for node in children]

        result = self
        if update:
            result = self.with_nodes(nodes)

        return result._replace(structure=result._structure.with_children(_structure, force=force))

    def with_node_parent(
            self, child: Union[str, AbstractNode], parent: Union[str, AbstractNode],
            *, force: bool = False, update: bool = False
    ) -> Self:
        result = self

        if update:
            nodes = {node for node in (child, parent) if isinstance(node, AbstractNode)}
            result = result.with_nodes(nodes)

        return result._replace(structure=result._structure.with_parent(_to_id(parent), _to_id(child), force=force))

    def with_roots(self, nodes: Iterable[Union[str, AbstractNode]], *, update: bool = False) -> Self:
        result = self

        if update:
            nodes = tuple(nodes)
            result = result.with_nodes({node for node in nodes if isinstance(node, AbstractNode)})

        return result._replace(structure=self._structure.as_roots(map(_to_id, nodes)))

    # semantic link methods

    @property
    def id2node_link(self) -> Dict[str, AbstractNodeLink]:
        return self.id2element(AbstractNodeLink)

    @property
    def node_links(self) -> Dict[Type[AbstractNodeLink], Iterable[AbstractNodeLink]]:
        return self.elements(AbstractNodeLink)

    def get_node_link(self, id_: str) -> AbstractNodeLink:
        return self.get_element(AbstractNodeLink, id_)

    def get_node_links(
            self, type_: Type[_NodeLink] = AbstractNodeLink, *,
            filter_: Union[Callable[[_NodeLink], bool], Iterable[Callable[[_NodeLink], bool]]] = tuple()
    ) -> Iterator[_NodeLink]:
        return self.get_elements(AbstractNodeLink, type_, filter_=filter_)

    def related_node_links(
            self, obj: Union[Identifiable, str], type_: Type[_NodeLink] = AbstractNodeLink, *,
            filter_: Union[Callable[[_NodeLink], bool], Iterable[Callable[[_NodeLink], bool]]] = tuple()
    ) -> Iterator[_NodeLink]:
        return self.related_elements(AbstractNodeLink, obj, type_, filter_=filter_)

    @_filter_elements(AbstractNodeLink)
    def with_node_links(self, links: Iterable[AbstractNodeLink], *, update: bool = False) -> Self:
        return self.with_elements(links, update=update)

    @_filter_elements(AbstractNodeLink)
    def without_node_links(self, links: Iterable[Union[str, AbstractNodeLink]], *, cascade: bool = False) -> Self:
        return self.without_elements(map(_to_id, links), cascade=cascade)

    # facts methods

    @property
    def id2fact(self) -> Dict[str, AbstractFact]:
        return self.id2element(AbstractFact)

    @property
    def facts(self) -> Dict[Type[AbstractFact], Iterable[AbstractFact]]:
        return self.elements(AbstractFact)

    def get_fact(self, id_: str) -> AbstractFact:
        return self.get_element(AbstractFact, id_)

    def get_facts(
            self, type_: Type[_Fact] = AbstractFact, *,
            filter_: Union[Callable[[_Fact], bool], Iterable[Callable[[_Fact], bool]]] = tuple()
    ) -> Iterator[_Fact]:
        return self.get_elements(AbstractFact, type_, filter_=filter_)

    def related_facts(
            self, obj: Union[Identifiable, str], type_: Type[_Fact] = AbstractFact, *,
            filter_: Union[Callable[[_Fact], bool], Iterable[Callable[[_Fact], bool]]] = tuple()
    ) -> Iterator[_Fact]:
        return self.related_elements(AbstractFact, obj, type_, filter_=filter_)

    @_filter_elements(AbstractFact)
    def with_facts(self, facts: Iterable[AbstractFact], *, update: bool = False) -> Self:
        return self.with_elements(facts, update=update)

    @_filter_elements(AbstractFact)
    def without_facts(self, facts: Iterable[Union[str, AbstractFact]], *, cascade: bool = False) -> Self:
        return self.without_elements(map(_to_id, facts), cascade=cascade)

    def __hash__(self):
        return hash((self._id, self._containers[AbstractNode], self._containers[AbstractFact]))

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, TalismanDocumentImpl):
            return NotImplemented
        return self._id == o._id and self._id2view == o._id2view and self._containers == o._containers and \
            self._structure == o._structure


def _to_id(obj: Union[str, EnsureIdentifiable]) -> str:
    return obj if isinstance(obj, str) else obj.id


def _filter(
        container: TypedIdsContainer[_D], known_ids: Dict[str, Any], elements: Iterable[Union[str, _D]], base_type: Type[_D]
) -> Iterator[Union[str, _D]]:
    for element in elements:
        if isinstance(element, str):
            if element in known_ids and element not in container:
                raise ValueError(f'Unexpected type for element {element}. Expected: {base_type}')
        else:
            if not isinstance(element, base_type):
                raise ValueError(f'Unexpected type for element {element}. Expected: {base_type}')
            if element.id in known_ids and element.id not in container:
                raise ValueError(f'Identifier collision for {element}')
        yield element



from typing import Type

from tdm.abstract.datamodel import AbstractNodeLink
from tdm.abstract.json_schema import AbstractLabeledModel, get_model_generator


def register_node_link_models() -> Type[AbstractLabeledModel[AbstractNodeLink]]:
    import tdm.datamodel.node_links as node_links
    node_links

    return get_model_generator(AbstractNodeLink).generate_labeled_model('NodeLinksModel')


NodeLinksModel = register_node_link_models()



from abc import ABCMeta
from dataclasses import dataclass
from typing import Generic, Optional, Sequence, Set, Tuple, TypeVar, Union

from tdm.abstract.datamodel.domain import AbstractValueDomainType
from tdm.abstract.datamodel.value import EnsureConfidenced
from tdm.helper import generics_mapping, is_subclass, unfold_union
from ._fact import AbstractFact, FactStatus

_V = TypeVar('_V', bound=EnsureConfidenced)
_DT = TypeVar('_DT', bound=AbstractValueDomainType)


@dataclass(frozen=True)
class AbstractValueFact(AbstractFact, Generic[_DT, _V], metaclass=ABCMeta):
    """
    Base class for value facts that represent some KB values.

    If the fact status is FactStatus.NEW `value` should be a tuple.
    If the fact status is FactStatus.APPROVED `value` should be single-valued.

    Attributes
    ----------
    type_id:
        The fact type identifier or instance of the value domain type.
    value:
        The normalized value or tuple of values (sorted by confidence) associated with the value fact.
    """
    type_id: Union[str, _DT]
    value: Union[Tuple[_V, ...], _V] = tuple()

    @property
    def str_type_id(self) -> str:
        return self.type_id if isinstance(self.type_id, str) else self.type_id.id

    @property
    def most_confident_value(self) -> Optional[_V]:
        if not self.value:
            return None
        if isinstance(self.value, Tuple):
            return self.value[0]
        return self.value

    def __post_init__(self):
        mapping = generics_mapping(type(self))
        domain_type = mapping.get(_DT)
        value_type = mapping.get(_V)
        if not issubclass(domain_type, AbstractValueDomainType):
            raise ValueError
        if not is_subclass(value_type, domain_type.get_value_type()):
            raise ValueError
        if not isinstance(self.type_id, str):
            if not isinstance(self.type_id, domain_type):
                raise ValueError(f"Illegal type id for fact {self}. {domain_type} is expected")
            value_type = self.type_id.value_type
        value_type = unfold_union(value_type)

        if isinstance(self.value, Sequence) and not isinstance(self.value, value_type):  # value could implement Sequence?
            if any(not isinstance(v, value_type) for v in self.value):
                raise ValueError(f"Fact {self} value should be {value_type} or tuple of {value_type}")
            object.__setattr__(self, 'value', tuple(self.value))
        elif not isinstance(self.value, value_type):
            raise ValueError(f"Fact {self} value should be {value_type} or tuple of {value_type}")

        if self.status is FactStatus.NEW and isinstance(self.value, value_type):
            object.__setattr__(self, 'value', (self.value,))
        elif self.status is FactStatus.APPROVED and isinstance(self.value, tuple) and not isinstance(self.value, value_type):
            if len(self.value) != 1 or not isinstance(self.value[0], value_type):
                raise ValueError(f"Approved fact {self} should have single value")
            object.__setattr__(self, 'value', self.value[0])

        if isinstance(self.value, Tuple):
            if all(v.confidence is not None for v in self.value):
                object.__setattr__(self, 'value', tuple(sorted(self.value, key=lambda v: v.confidence, reverse=True)))
            elif not all(v.confidence is None for v in self.value):
                raise ValueError(f"Fact {self} couldn't contain values with confidence and values without confidence at same time")

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}



__all__ = [
    'AbstractDomain',
    'AbstractLinkDomainType',
    'AbstractDomainType',
    'AbstractValueDomainType'
]

from ._domain import AbstractDomain
from ._link import AbstractLinkDomainType
from ._type import AbstractDomainType
from ._value import AbstractValueDomainType



from dataclasses import dataclass
from typing import Generic, TypeVar, Union

from tdm.abstract.datamodel.value import AbstractValue, EnsureConfidenced
from tdm.abstract.json_schema import generate_model

_S = TypeVar('_S')


@dataclass(frozen=True)
class _ScalarValue(EnsureConfidenced, Generic[_S]):
    """
    Represents the internal structure of a scalar value.

    Attributes
    ----------
    value:
        Normalized scalar value
    """
    value: _S

    def __post_init__(self):
        if type(self) is _ScalarValue:
            raise TypeError


@generate_model
@dataclass(frozen=True)
class StringValue(AbstractValue, _ScalarValue[str]):
    """
    Represents a normalized KB value for strings.
    """
    pass


@generate_model
@dataclass(frozen=True)
class IntValue(AbstractValue, _ScalarValue[int]):
    """
    Represents a normalized KB value for integers.
    """
    pass


@generate_model
@dataclass(frozen=True)
class DoubleValue(AbstractValue, _ScalarValue[Union[float, int]]):
    """
    Represents a normalized KB value for float values.
    """
    pass


@generate_model
@dataclass(frozen=True)
class TimestampValue(AbstractValue, _ScalarValue[int]):
    """
    Represents a normalized KB value for unix timestamps.
    """
    pass



from typing import Type

from tdm.abstract.datamodel import AbstractFact
from tdm.abstract.json_schema import AbstractLabeledModel, get_model_generator


def register_fact_models() -> Type[AbstractLabeledModel[AbstractFact]]:
    import tdm.datamodel.facts as facts  # we need this import for serializers registration
    facts

    # TODO: here plugin for extra document nodes could be added

    return get_model_generator(AbstractFact).generate_labeled_model('FactsModel')


FactsModel = register_fact_models()



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from enum import Enum
from functools import total_ordering
from typing import Callable, Iterable, Type, Union

from typing_extensions import Self

from tdm.abstract.datamodel.domain import AbstractDomain
from tdm.abstract.datamodel.identifiable import EnsureIdentifiable


@total_ordering
class FactStatus(str, Enum):
    """
    Enumeration class representing the status of a fact.

    Attributes
    ----------
    APPROVED:
        Fact is already approved to be correct and already stored in KB.
    DECLINED:
        Fact is already rejected (incorrect fact).
    AUTO:
        Fact is marked to be approved automatically (correct fact not stored in KB yet).
    HIDDEN:
        Fact is neither approved nor declined, but is not relevant for downstream task.
    NEW:
        Fact is neither approved nor declined.
    """

    def __new__(cls, name: str, priority: int):
        obj = str.__new__(cls, name)
        obj._value_ = name
        obj.priority = priority
        return obj

    APPROVED = ("approved", 0)
    DECLINED = ("declined", 1)
    AUTO = ("auto", 2)
    HIDDEN = ("hidden", 3)
    NEW = ("new", 4)

    def __lt__(self, other: 'FactStatus'):
        if not isinstance(other, FactStatus):
            return NotImplemented
        return self.priority < other.priority


@dataclass(frozen=True)
class AbstractFact(EnsureIdentifiable, metaclass=ABCMeta):
    """
    The most abstract class for facts.

    Attributes
    ----------
    status:
        Status of the fact
    """
    status: FactStatus

    @abstractmethod
    def replace_with_domain(self, domain: AbstractDomain) -> Self:
        """
        Create a new fact with domain type instead of its identifier.
        This method also validates domain type restrictions.

        :param domain: domain to be used for fact validation.
        :return: New fact object with domain type instead of its identifier.
        """
        pass

    @staticmethod
    def id_filter(obj: Union['AbstractFact', str]) -> Callable[['AbstractFact'], bool]:
        """
        Generate a filter function to filter facts based on their identifiers.

        :param obj: The fact object or identifier to filter.
        :return: The filter function that returns True for facts with a matching identifier.
        """
        id_ = obj.id if isinstance(obj, AbstractFact) else obj

        def _filter(fact: AbstractFact) -> bool:
            return fact.id == id_

        return _filter

    @staticmethod
    def status_filter(status: Union[FactStatus, Iterable[FactStatus]]) -> Callable[['AbstractFact'], bool]:
        """
        Generate a filter function to filter facts based on their status or a collection of statuses.

        :param status: The status or collection of statuses to filter.
        :return: The filter function that returns True for facts with a matching status.
        """
        if isinstance(status, FactStatus):
            def _filter(fact: AbstractFact) -> bool:
                return fact.status is status
        else:
            statuses = frozenset(status)

            def _filter(fact: AbstractFact) -> bool:
                return fact.status in statuses

        return _filter

    @staticmethod
    def type_filter(type_: Type['AbstractFact']) -> Callable[['AbstractFact'], bool]:
        """
        Generate a filter function to filter facts based on their type.

        :param type_: The type of facts to filter.
        :return: The filter function that returns True for facts of the specified type.
        """

        def _filter(fact: AbstractFact) -> bool:
            return isinstance(fact, type_)

        return _filter



from abc import ABCMeta
from dataclasses import dataclass

from .node import AbstractNode


@dataclass(frozen=True)
class AbstractNodeMention(metaclass=ABCMeta):
    """
    Abstract class for node mentions.

    Attributes
    --------
    node:
        mentioned node
    """
    node: AbstractNode

    @property
    def node_id(self) -> str:
        """
        :return: mentioned node identifier
        """
        return self.node.id



import dataclasses
from dataclasses import fields
from typing import Any, Dict, Generic, Optional, Type, TypeVar

from pydantic import BaseModel, Extra, Field, create_model, root_validator
from typing_extensions import Literal, Self

from tdm.helper import cache_result, generics_mapping, register_in_module
from .serializers import AbstractElementModel, AbstractElementSerializer, get_serializer

_Element = TypeVar('_Element')


# hackish root validator solution as normal validators are not called for skipped required fields
def set_type_if_none(cls, values):
    if 'type' not in values:
        values['type'] = cls.__fields__['type'].type_.__args__[0]  # get arg of Literal ('type' field type)
    return values


class ElementModel(BaseModel, AbstractElementModel[_Element], Generic[_Element]):
    class Config:
        extra = Extra.forbid

        @staticmethod
        def schema_extra(schema: Dict[str, Any]):
            for prop in schema.get('properties', {}).values():
                prop.pop('_serializer', None)

    def deserialize(self, typed_id2element: Dict[type, Dict[str, Any]]) -> _Element:
        raise NotImplementedError

    @classmethod
    def serialize(cls, element: _Element) -> Self:
        kwargs = {}
        for key, value in element.__dict__.items():
            kwargs[key] = cls.__fields__[key].field_info.extra['_serializer'].serialize(value)
        return cls(**kwargs)


def _wrap_deserializer(coll_type: type, serializer: AbstractElementSerializer):
    def deserialize(values, *args):
        return coll_type(serializer.deserialize(v, *args) for v in values)

    return deserialize


_BaseType = TypeVar('_BaseType', bound=type)


@cache_result()
def create_model_for_type(type_: Type[_Element], label: Optional[str] = None) -> Type[ElementModel[_Element]]:
    type_vars = generics_mapping(type_)

    model_fields = {}
    validators = {}

    for field in fields(type_):
        name = field.name
        default_value = field.default if field.default is not dataclasses.MISSING else ...

        field_type, serializer = get_serializer(field.type, type_vars)

        if hasattr(field_type, '__metadata__'):  # field type is already Annotated
            metadata = field_type.__metadata__
            if isinstance(metadata, tuple):
                metadata = metadata[0]
            discriminator = metadata.discriminator
            model_fields[name] = (
                field_type.__origin__,
                Field(default_value, discriminator=discriminator, _serializer=serializer)
            )
        else:
            model_fields[name] = (field_type, Field(default_value, _serializer=serializer))
    if label:
        model_fields['type'] = (Literal[label], ...)
        validators['set_if_none'] = root_validator(pre=True, allow_reuse=True)(set_type_if_none)

    model = register_in_module(create_model(f"{type_.__name__}Model", __base__=ElementModel, __validators__=validators, **model_fields))

    def deserialize(self: model, typed_id2element: Dict[type, Dict[str, Any]]) -> type_:
        kwargs = {}
        for f in set(type_.__dataclass_fields__).intersection(self.__dict__):
            kwargs[f] = self.__fields__[f].field_info.extra['_serializer'].deserialize(getattr(self, f), typed_id2element)
        return type_(**kwargs)

    model.deserialize = deserialize

    return model



from typing import Callable, TypeVar

from .fact import AbstractFact

_FT = TypeVar('_FT', bound=AbstractFact)


def and_filter(*args: Callable[[_FT], bool]) -> Callable[[_FT], bool]:
    def _filter(fact: _FT) -> bool:
        return all(f(fact) for f in args)

    return _filter


def or_filter(*args: Callable[[_FT], bool]) -> Callable[[_FT], bool]:
    def _filter(fact: _FT) -> bool:
        return any(f(fact) for f in args)

    return _filter


def not_filter(f: Callable[[_FT], bool]) -> Callable[[_FT], bool]:
    def _filter(fact: _FT) -> bool:
        return not f(fact)

    return _filter



from typing import Callable, Dict, Tuple, Type

from tdm.abstract.datamodel.value import AbstractValue
from tdm.abstract.json_schema import ElementModel, get_model_generator


def register_value_models() -> Tuple[Dict[Type[AbstractValue], Type[ElementModel]], Callable[[AbstractValue], ElementModel]]:
    import tdm.datamodel.values as values
    values

    # TODO: add plugins here

    return get_model_generator(AbstractValue).generate_union_model()


VALUE_MODELS, serialize_value = register_value_models()



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractNodeLink, AbstractNodeMention, Identifiable
from tdm.abstract.json_schema import generate_model


@generate_model(label='reference')
@dataclass(frozen=True)
class ReferenceNodeLink(Identifiable, AbstractNodeLink[AbstractNodeMention, AbstractNodeMention]):
    """
    Represents a reference link between two node mentions (footnote, bibliography, etc.).
    """
    pass



from collections import defaultdict
from typing import Callable, Dict, List, Tuple

from tdm.abstract.datamodel import AbstractNode, AbstractNodeLink, BaseNodeMetadata, FrozenMarkup
from tdm.datamodel.mentions import NodeMention
from tdm.datamodel.node_links import TranslationNodeLink
from tdm.datamodel.nodes import ImageNode, ImageNodeMetadata, JSONNode, KeyNode, ListNode, ListNodeMetadata, TableCellNode, \
    TableNode, TableRowNode, TextNode, TextNodeMetadata
from tdm.v0.json_schema.content import NodeType, TreeDocumentContentModel


def _convert_text_node(node: TreeDocumentContentModel) -> TextNode:
    return TextNode(
        content=node.text,
        markup=FrozenMarkup.freeze(node.markup.dict()),
        id=node.id,
        metadata=TextNodeMetadata(hidden=node.metadata.hidden, language=node.metadata.language)
    )


def _covert_key_node(node: TreeDocumentContentModel) -> KeyNode:
    return KeyNode(
        content=node.text,
        markup=FrozenMarkup.freeze(node.markup.dict()),
        id=node.id,
        metadata=TextNodeMetadata(hidden=node.metadata.hidden, language=node.metadata.language)
    )


def _convert_list_node(node: TreeDocumentContentModel) -> ListNode:
    return ListNode(id=node.id, metadata=ListNodeMetadata(hidden=node.metadata.hidden))


def _convert_json_node(node: TreeDocumentContentModel) -> JSONNode:
    return JSONNode(id=node.id, metadata=BaseNodeMetadata(hidden=node.metadata.hidden))


def _convert_table_node(node: TreeDocumentContentModel) -> TableNode:
    return TableNode(id=node.id, metadata=BaseNodeMetadata(hidden=node.metadata.hidden))


def _convert_table_row_node(node: TreeDocumentContentModel) -> TableRowNode:
    return TableRowNode(id=node.id, metadata=BaseNodeMetadata(hidden=node.metadata.hidden))


def _convert_image_node(node: TreeDocumentContentModel) -> ImageNode:
    return ImageNode(
        content=node.text,
        markup=FrozenMarkup.freeze(node.markup.dict()),
        id=node.id,
        metadata=ImageNodeMetadata(hidden=node.metadata.hidden)
    )


NODE_CONVERTERS: Dict[NodeType, Callable[[TreeDocumentContentModel], AbstractNode]] = {
    NodeType.HEADER: _convert_text_node,
    NodeType.TEXT: _convert_text_node,
    NodeType.LIST: _convert_list_node,
    NodeType.JSON: _convert_json_node,
    NodeType.KEY: _covert_key_node,
    NodeType.TABLE: _convert_table_node,
    NodeType.TABLE_ROW: _convert_table_row_node,
    NodeType.IMAGE: _convert_image_node
}


def build_structure(root: TreeDocumentContentModel) -> Tuple[List[AbstractNode], Dict[str, List[str]], List[AbstractNodeLink]]:
    nodes: list[AbstractNode] = []
    structure: dict[str, list[str]] = defaultdict(list)
    node_links: list[AbstractNodeLink] = []

    to_process = [root]  # old document is single tree, so we can start from root

    for node in to_process:
        # create new node for each document node with same type
        converted = NODE_CONVERTERS[node.metadata.node_type](node)
        nodes.append(converted)

        # create separate node with original text (current node is target, original text is source), type is the same
        if node.metadata.original_text is not None:
            if node.metadata.node_type is not NodeType.KEY:
                raise ValueError
            node_links.append(TranslationNodeLink(
                source=NodeMention(KeyNode(content=node.metadata.original_text)),
                target=NodeMention(converted)
            ))

        # text translations are moved to separate nodes linked by translation link (current node is source, translation is target)
        if node.metadata.text_translations:
            if node.metadata.node_type is not NodeType.TEXT and node.metadata.node_type is not NodeType.HEADER:
                raise ValueError
            node_links.extend(
                TranslationNodeLink(
                    source=NodeMention(converted),
                    target=NodeMention(TextNode(content=text, metadata=TextNodeMetadata(language=lang)))
                ) for lang, text in node.metadata.text_translations.items()
            )

        children = node.nodes
        if children:
            if isinstance(converted, TableRowNode):
                # table row should contain only cell nodes (there are no such nodes in old document)
                for c in children:
                    cell = TableCellNode()
                    nodes.append(cell)
                    structure[converted.id].append(cell.id)
                    structure[cell.id].append(c.id)
            else:
                structure[converted.id].extend(map(lambda n: n.id, children))
            to_process.extend(children)

    return nodes, structure, node_links



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractValue, Identifiable
from tdm.abstract.datamodel.domain import AbstractValueDomainType


@dataclass(frozen=True)
class AtomValueType(Identifiable, AbstractValueDomainType[AbstractValue]):
    """
    Domain type for atom value facts
    """
    pass



__all__ = [
    'CompositeValueType',
    'AbstractConceptType', 'AccountType', 'ConceptType', 'DocumentType', 'PlatformType',
    'PropertyType', 'RelationPropertyType',
    'RelationType', 'ComponentValueType', 'AtomValueType'
]

from ._component import ComponentValueType
from ._composite import CompositeValueType
from ._concept import AbstractConceptType, AccountType, ConceptType, DocumentType, PlatformType
from ._property import PropertyType, RelationPropertyType
from ._relation import RelationType
from ._value import AtomValueType



from collections import defaultdict
from typing import ClassVar, Dict, Iterable, Optional, Tuple, Type, TypeVar, Union

from pydantic import BaseModel, PrivateAttr

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, TalismanDocument
from tdm.datamodel.document import TalismanDocumentFactory
from tdm.datamodel.domain import PropertyType
from tdm.datamodel.domain.types import AbstractConceptType, AccountType, DocumentType, PlatformType
from .content import TreeDocumentContentModel
from .convert import build_structure, convert_account_directive, convert_concept_fact, convert_platform_directive, convert_property_fact, \
    convert_relation_fact, convert_value_fact, get_metadata_facts
from .directive import DirectiveModel, DirectiveType
from .fact import ConceptFactModel, FactModel, FactType, PropertyFactModel, RelationFactModel, ValueFactModel
from .metadata import DocumentMetadataModel

_T = TypeVar('_T', bound=AbstractConceptType)


def _get_source_type(domain: AbstractDomain, cls: Type[_T], props: Iterable[PropertyType]) -> _T:
    sources = {domain.get_type(p.id).source for p in props}
    if len(sources) > 1:
        raise ValueError
    source = sources.pop()
    if not isinstance(source, cls):
        raise ValueError
    return source


class TalismanDocumentModel(BaseModel):
    _DEFAULT_DOMAIN: ClassVar[AbstractDomain] = None
    _domain: AbstractDomain = PrivateAttr(None)

    id: str
    content: TreeDocumentContentModel
    metadata: Optional[DocumentMetadataModel]
    facts: Optional[Tuple[FactModel, ...]]  # pydantic fails to serialize FrozenSet[FactModel]
    directives: Optional[Tuple[DirectiveModel, ...]]

    @classmethod
    def set_default_domain(cls, domain: Optional[AbstractDomain]) -> None:
        cls._DEFAULT_DOMAIN = domain

    def set_domain(self, domain: Optional[AbstractDomain]) -> None:
        self._domain = domain

    def to_doc(
            self, doc_type: Union[str, DocumentType], concept_titles: Iterable[PropertyType], metadata_map: Dict[str, str],
            platform_props: Dict[str, PropertyType] = None, account_props: Dict[str, PropertyType] = None
    ) -> TalismanDocument:
        domain = self._domain or self._DEFAULT_DOMAIN
        if domain is None:
            raise ValueError('TDM v0 could parse document only with domain')
        titles: Dict[str, PropertyType] = {prop.source.id: prop for prop in concept_titles}

        result = TalismanDocumentFactory(self._domain or self._DEFAULT_DOMAIN).create_document(id_=self.id, doc_type=doc_type)

        nodes, structure, node_links = build_structure(self.content)
        result = result.with_nodes(nodes).with_structure(structure).with_node_links(node_links, update=True).with_main_root(self.content.id)

        # convert document metadata to facts
        if self.metadata:
            result = result.with_facts(get_metadata_facts(result, self.metadata.to_metadata(), self._domain, metadata_map), update=True)

        # now convert facts
        old_facts = defaultdict(set)
        for fact in (self.facts or []):
            old_facts[fact.fact_type].add(fact)

        facts = []
        facts_mapping: dict[str, AbstractFact] = {}
        # convert all CreateAccount and CreateAccount directives into new types
        type2directives = defaultdict(list)
        for d in self.directives or ():
            type2directives[d.directive_type].append(d)
        if DirectiveType.CREATE_PLATFORM in type2directives:
            platform_type = _get_source_type(domain, PlatformType, platform_props.values())
            for d in type2directives[DirectiveType.CREATE_PLATFORM]:
                facts.extend(convert_platform_directive(d, platform_type, platform_props))
        if DirectiveType.CREATE_ACCOUNT in type2directives:
            account_type = _get_source_type(domain, AccountType, account_props.values())
            for d in type2directives[DirectiveType.CREATE_ACCOUNT]:
                facts.extend(convert_account_directive(d, account_type, account_props))

        # start with concept facts
        # old concept fact consists of <status, type_id, value, mention> and should be converted to chain
        # ConceptFact(type_id) <- Property(Name) -> AtomValueFact(value) <-  MentionFact -> TextNodeMention(mention)

        for old_concept_fact in old_facts.get(FactType.CONCEPT, []):
            old_concept_fact: ConceptFactModel
            concept_fact, *other = convert_concept_fact(old_concept_fact, result, domain, titles)
            facts_mapping[old_concept_fact.id] = concept_fact
            facts.extend((concept_fact, *other))
        # then convert ValueFacts
        # old value fact consists of <status, type_id, value, mention> and should be converted to chain
        # AtomValueFact(value) <- MentionFact -> TextNodeMention(mention)
        for old_value_fact in old_facts.get(FactType.VALUE, []):
            old_value_fact: ValueFactModel
            value_fact, *other = convert_value_fact(old_value_fact, result, domain)
            facts_mapping[old_value_fact.id] = value_fact
            facts.extend((value_fact, *other))

        # now relations
        # old relations should be mapped to new ones with trivial modifications
        for old_relation_fact in old_facts.get(FactType.RELATION, []):
            old_relation_fact: RelationFactModel
            relation_fact = convert_relation_fact(old_relation_fact, facts_mapping)
            facts_mapping[old_relation_fact.id] = relation_fact
            facts.append(relation_fact)

        # and finally properties and relation properties
        for old_property_fact in old_facts.get(FactType.PROPERTY, []):
            old_property_fact: PropertyFactModel
            property_fact = convert_property_fact(old_property_fact, facts_mapping)
            facts.append(property_fact)

        result = result.with_facts(facts, update=True)
        return result



from abc import abstractmethod

from immutabledict import immutabledict
from typing_extensions import Self


class AbstractMarkup(object):
    """
    An abstract base class for custom node markup creation and manipulation.

    All the implementations should be immutable.
    """
    @property
    @abstractmethod
    def markup(self) -> immutabledict:
        """
        Retrieve the current markup associated with the instance.

        This property is utilized for equality check.

        :return: The current markup (immutable)
        """
        pass

    @classmethod
    @abstractmethod
    def from_markup(cls, markup: 'AbstractMarkup') -> Self:
        """
        Create a new instance of the node markup based on an existing markup.

        :param markup: The existing markup to base the new instance on.
        :return: A new instance of the node markup class.
        """
        pass

    def __hash__(self):
        return hash(self.markup)

    def __eq__(self, other):
        if not isinstance(other, AbstractMarkup):
            return NotImplemented
        return self.markup == other.markup



from dataclasses import dataclass
from typing import Optional

from tdm.abstract.datamodel import AbstractNode, BaseNodeMetadata
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class ListNodeMetadata(BaseNodeMetadata):
    """
    List node metadata

    Attributes
    --------
    bullet:
        bullet type used for each list item
    """
    bullet: Optional[str] = None


@generate_model(label='list')
@dataclass(frozen=True)
class ListNode(AbstractNode[ListNodeMetadata]):
    """
    Service node to represent ordered list.
    Each direct child is an item of this list.
    """
    pass


@generate_model(label='json')
@dataclass(frozen=True)
class JSONNode(AbstractNode[BaseNodeMetadata]):
    """
    Service node to represent key-value mapping.
    Children nodes are keys in key-value pairs.

    Each direct child should be ``KeyNode``.
    """
    pass


@generate_model(label='table')
@dataclass(frozen=True)
class TableNode(AbstractNode[BaseNodeMetadata]):
    """
    Service node to represent table.
    Children nodes are rows of the table.

    Each direct child should be ``TableRowNode``.
    """
    pass


@generate_model(label='row')
@dataclass(frozen=True)
class TableRowNode(AbstractNode[BaseNodeMetadata]):
    """
    Service node to represent table row.
    Children nodes ore cells of the table row.

    Each direct child should be ``TableCell``.
    Nodes of this type could be direct children only for a ``TableNode``
    """
    pass


@dataclass(frozen=True)
class TableCellNodeMetadata(BaseNodeMetadata):
    """
    Table cell metadata.

    Attributes
    --------
    header:
        header indicator
    colspan:
        width of the cell (in columns)
    rowspan:
        height of the cell (in rows)
    """
    header: Optional[bool] = None
    colspan: Optional[int] = None
    rowspan: Optional[int] = None


@generate_model(label='cell')
@dataclass(frozen=True)
class TableCellNode(AbstractNode[TableCellNodeMetadata]):
    """
    Service node to represent table cell.
    Children nodes ore the cell content.

    Nodes of this type could be direct children only for a ``TableRowNode``
    """
    pass



import uuid
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Set


@dataclass(frozen=True)
class EnsureIdentifiable(object):  # we need such ugly inheritance to guarantee default valued fields follows fields without defaults
    """
    Base interface for identifiable objects that contains non-default fields.
    Python dataclasses before 3.10 couldn't define kwargs-only fields and ``Identifiable`` class contains one (id).

    In cases identifiable object contains non-default fields one should create intermediate ``class _MyClass(EnsureIdentifiable)`` where
    fields are defined.
    Final class will inherit both classes ``class MyClass(Identifiable, _MyClass)``
    """

    def __post_init__(self):
        if not isinstance(self, Identifiable):
            raise TypeError(f"{type(self)} should inherit {Identifiable}. Actual mro is {type(self).mro()}")

    @classmethod
    @abstractmethod
    def constant_fields(cls) -> Set[str]:
        """
        Get set of fields that are constant for identifiable object.
        Some identifiable objects could depend on other objects (and values stored in these objects).
        So changing these fields could lead to document inconsistency.
        If some constant fields should be changed, remove objects from document and then add any desired objects.

        :return: set of fields that should be never changed
        """
        pass


@dataclass(frozen=True, eq=False)
class Identifiable(EnsureIdentifiable, metaclass=ABCMeta):
    """
    Base interface for identifiable objects that could be stored in document.

    This class should be the first class with ``__post_init__`` method implemented in inherited class mro to guarantee correct identifier
    generation.
    In other cases ``Identifiable.__post_init__(self)`` should be called explicitly.

    Attributes
    --------
    id:
        Object unique identifier. It will be automatically generated if no passed to object constructor
    """
    id: str = field(default_factory=lambda: None)

    def __post_init__(self):
        if self.id is None:
            object.__setattr__(self, 'id', self.generate_id())
        for type_ in type(self).mro():
            if issubclass(type_, Identifiable):
                continue
            if hasattr(type_, '__post_init__'):
                type_.__post_init__(self)

    @staticmethod
    def generate_id() -> str:
        return str(uuid.uuid4())



import importlib
from typing import Any, Dict, Generic, Type, TypeVar

from .abstract import AbstractElementSerializer

_ARG = TypeVar('_ARG')


class TypeSerializer(AbstractElementSerializer[Type[_ARG], str], Generic[_ARG]):
    def serialize(self, element: Type[_ARG]) -> str:
        module = element.__module__
        if module == 'builtins':
            return element.__qualname__
        return f"{module}.{element.__qualname__}"

    def deserialize(self, serialized: str, typed_id2element: Dict[type, Dict[str, Any]]) -> Type[_ARG]:
        if '.' not in serialized:
            import builtins
            return getattr(builtins, serialized)
        module_name, name = serialized.rsplit('.', 1)
        module = importlib.import_module(module_name)
        return getattr(module, name)

    def field_type(self, element_type: Type[_ARG]) -> Type[str]:
        return str



from dataclasses import dataclass
from typing import Optional, Union

from tdm.abstract.datamodel.value import AbstractValue, EnsureConfidenced
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class Coordinates(object):
    latitude: Union[float, int]
    longitude: Union[float, int]


@dataclass(frozen=True)
class _GeoPointValue(EnsureConfidenced):
    """
    Auxiliary class for `GeoPointValue` to fix dataclass fields order.

    Represents the internal structure of a `GeoPointValue`.

    Attributes
    ----------
    point:
        Optional normalized geo coordinates representation.
    name:
        Optional normalized location name representation.
    """
    point: Optional[Coordinates] = None
    name: Optional[str] = None


@generate_model
@dataclass(frozen=True)
class GeoPointValue(AbstractValue, _GeoPointValue):
    """
    Represents a normalized KB value for geolocation.
    """

    @classmethod
    def from_dict(cls, value: dict) -> 'GeoPointValue':
        """
        Create a `GeoPointValue` object from a dictionary.

        Expected value scheme is as follows:

        {
          "point": {
            "latitude": float,
            "longitude": float
          },
          "name": str
        }

        Both parts could be skipped.

        :param value: The dictionary containing the GeoPointValue information.
        :return: A GeoPointValue object.
        """
        args = {}
        if 'point' in value:
            args['point'] = Coordinates(**value['point'])
        if 'name' in value:
            args['name'] = value['name']
        return cls(**args)



from typing import Mapping, Sequence

from immutabledict import immutabledict
from pydantic import BaseModel


def freeze_dict(obj: Mapping) -> immutabledict:
    return immutabledict((k, freeze(v)) for k, v in obj.items())


def freeze_sequence(obj: Sequence) -> tuple:
    return tuple(freeze(v) for v in obj)


def freeze(obj):
    if obj is None or isinstance(obj, (str, int, float, bool)):
        return obj
    if isinstance(obj, Mapping):
        return freeze_dict(obj)
    if isinstance(obj, set):
        return frozenset(freeze(v) for v in obj)
    if isinstance(obj, Sequence):
        return freeze_sequence(obj)
    raise ValueError


def unfreeze(obj):
    if obj is None or isinstance(obj, (str, int, float, bool)):
        return obj
    if isinstance(obj, Mapping):
        return {k: unfreeze(v) for k, v in obj.items()}
    if isinstance(obj, (frozenset, set)):
        return {unfreeze(v) for v in obj}
    if isinstance(obj, Sequence):
        t = type(obj)
        return t(unfreeze(v) for v in obj)
    if isinstance(obj, BaseModel):
        return unfreeze(obj.dict())



from collections import defaultdict
from itertools import chain
from typing import Callable, Dict, Iterable, Iterator, Set, Tuple, Type, TypeVar, Union

from typing_extensions import Self

from tdm.abstract.datamodel import EnsureIdentifiable, Identifiable, and_filter
from tdm.helper import unfold_union
from ._container import TypedIdsContainer
from ._state import pack_view_state, unpack_view_state
from ._types import get_base_type
from ._view import AbstractView, object_view, restore_object

_I = TypeVar('_I', bound=Identifiable)
_Base = TypeVar('_Base', bound=EnsureIdentifiable)


class ViewContainer(object):
    __slots__ = (
        '_id2view', '_dependencies', '_containers'
    )

    def __setstate__(self, state):
        d, s = state
        if d is not None:
            self.__dict__.update(d)
        s['_id2view'] = unpack_view_state(s['_id2view'])
        for component in chain.from_iterable(getattr(cls, '__slots__', ()) for cls in type(self).mro()):
            setattr(self, component, s.get(component, None))

    def __getstate__(self):
        result = {component: getattr(self, component, None) for component in
                  chain.from_iterable(getattr(cls, '__slots__', ()) for cls in type(self).mro())}
        result['_id2view'] = pack_view_state(result.get('_id2view', {}))
        d = self.__dict__ if hasattr(self, '__dict__') else None
        return d, result

    def __init__(
            self,
            id2view: Dict[str, Union[EnsureIdentifiable, AbstractView]],
            dependencies: Dict[str, Set[Tuple[str, Type[EnsureIdentifiable]]]],
            containers: Dict[Type[EnsureIdentifiable], TypedIdsContainer],
    ):
        self._id2view = id2view
        self._dependencies = dependencies
        self._containers = containers

    def _replace(self, **kwargs) -> Self:
        return type(self)(
            **{
                'id2view': self._id2view,
                'dependencies': self._dependencies,
                'containers': self._containers,
                **kwargs
            }
        )

    def id2element(self, base_type: Type[_Base]) -> Dict[str, _Base]:
        return {i: restore_object(self._id2view[i], self._id2view) for i, _ in self._containers[base_type]}

    def elements(self, base_type: Type[_Base]) -> Dict[Type[_Base], Iterable[_Base]]:
        return {
            t: {restore_object(self._id2view[i], self._id2view) for i in ids}
            for t, ids in self._containers[base_type].type2ids.items()
        }

    def get_element(self, base_type: Type[EnsureIdentifiable], id_: str) -> _I:
        if id_ not in self._containers[base_type]:
            raise KeyError
        return restore_object(self._id2view[id_], self._id2view)

    def get_elements(
            self, base_type: Type[EnsureIdentifiable], type_: Type[_I],
            filter_: Union[Callable[[_I], bool], Iterable[Callable[[_I], bool]]]
    ) -> Iterator[_I]:
        if isinstance(filter_, Iterable):
            filter_ = and_filter(*filter_)

        type_ = unfold_union(type_)
        if not all(issubclass(t, base_type) for t in type_):
            raise ValueError

        for t, ids in self._containers[base_type].type2ids.items():
            if not issubclass(t, type_):
                continue
            yield from filter(filter_, (restore_object(self._id2view[id_], self._id2view) for id_ in ids))

    def related_elements(
            self, base_type: Type[EnsureIdentifiable], obj: Union[Identifiable, str], type_: Type[_I],
            filter_: Union[Callable[[_I], bool], Iterable[Callable[[_I], bool]]]
    ) -> Iterator[_I]:
        if isinstance(obj, Identifiable):
            obj = obj.id
        if not isinstance(obj, str):
            raise ValueError

        if isinstance(filter_, Iterable):
            filter_ = and_filter(*filter_)

        type_ = unfold_union(type_)
        if not all(issubclass(t, base_type) for t in type_):
            raise ValueError

        for element_id, element_type in self._dependencies.get(obj, tuple()):
            if not issubclass(element_type, base_type):
                continue
            view = self._id2view[element_id]
            if not issubclass(view.orig_type(), type_):
                continue
            fact = restore_object(view, self._id2view)
            if filter_(fact):
                yield fact

    def with_elements(self, elements: Iterable[EnsureIdentifiable], *, update: bool = False) -> Self:
        containers: Dict[Type[EnsureIdentifiable], TypedIdsContainer] = dict(self._containers)
        id2view = dict(self._id2view)
        dependencies = defaultdict(set, self._dependencies)
        updated = set()

        elements = self._order_dependencies(set(elements), update)

        for element_type, element, element_view in elements:
            assert isinstance(element_view, AbstractView)
            container = containers[element_type]
            if element.id in container:
                try:
                    id2view[element.id].validate_update(element_view)
                except ValueError as e:
                    raise ValueError(f"Couldn't update element {element}") from e
            elif element.id in id2view:
                raise ValueError(f"Element {element} identifiers collision: document already "
                                 f"contains {id2view[element.id].orig_type()} with same id")
            if isinstance(element_view, AbstractView):
                for dep in element_view.__depends_on__:
                    if dep not in updated:
                        dependencies[dep] = set(dependencies[dep])
                        updated.add(dep)
                    dependencies[dep].add((element.id, element_type))
            id2view[element.id] = element_view
            containers[element_type] = containers[element_type].with_ids([(element.id, type(element))])

        return self._replace(
            id2view=id2view,
            dependencies=dependencies,
            containers=containers
        )

    def without_elements(self, ids: Iterable[str], *, cascade: bool = False) -> Self:
        ids = set(self._id2view.keys()).intersection(map(_to_id, ids))  # ignore excess ids
        if not cascade:
            # check no hang links
            for id_ in ids:
                for dep, _ in self._dependencies.get(id_, ()):
                    if dep not in ids:
                        raise ValueError(f"Couldn't remove element {id_} as it depends on element {dep}")
        remove_from_containers = defaultdict(set)
        id2view = dict(self._id2view)
        dependencies = dict(self._dependencies)
        updated = set()
        remove = defaultdict(set)
        removed_ids = set()
        for id_ in ids:
            remove[get_base_type(id2view[id_].orig_type())].add(id_)
        while remove:
            to_remove = defaultdict(set)
            for base_type, ids in remove.items():
                for id_ in ids:
                    if id_ in removed_ids:
                        continue
                    removed_ids.add(id_)
                    for dep_id, dep_type in dependencies.get(id_, ()):
                        to_remove[dep_type].add(dep_id)
                    view = id2view.pop(id_)
                    dependencies.pop(id_, None)
                    remove_from_containers[base_type].add(id_)
                    if not isinstance(view, AbstractView):
                        continue
                    for dep in view.__depends_on__:
                        if dep not in dependencies:
                            continue
                        if dep not in updated:
                            dependencies[dep] = set(dependencies[dep])
                            updated.add(dep)
                        dependencies[dep].discard((id_, base_type))
                        if not dependencies[dep]:
                            dependencies.pop(dep)
            remove = to_remove

        containers = dict(self._containers)
        for base_type, ids in remove_from_containers.items():
            containers[base_type] = containers[base_type].without_ids(ids)

        return self._replace(id2view=id2view, dependencies=dependencies, containers=containers)

    def _order_dependencies(
            self, elements: Set[EnsureIdentifiable], update: bool = False
    ) -> Iterable[Tuple[Type[EnsureIdentifiable], EnsureIdentifiable, AbstractView]]:
        result = []
        visited = set()
        while elements:
            to_process = set()
            for element in map(self._transform_element, elements):
                visited.add(element)
                view = object_view(element)

                result.append((get_base_type(element), element, view))

                if not isinstance(view, AbstractView):
                    continue  # no dependencies

                dependencies = view.get_dependencies(element)
                dependencies.difference_update(elements)  # will be processed next steps
                dependencies.difference_update(visited)
                if update:
                    to_process.update(dependencies)  # add to queue
                else:
                    self._validate_elements(dependencies)
            elements = to_process
        return result[::-1]

    def _transform_element(self, element: _Base) -> _Base:
        return element

    def _validate_elements(self, elements: Set[EnsureIdentifiable]) -> None:
        for element in elements:
            if element.id not in self._id2view:
                raise ValueError(f"document contains no {element}")
            view = self._id2view[element.id]
            element_type = type(element)
            view_type = view.orig_type() if isinstance(view, AbstractView) else type(view)
            if not issubclass(element_type, view_type) or not issubclass(view_type, element_type):
                raise ValueError(f"Type mismatch for {element}. Expected: {view_type}, actual: {element_type}")


def _to_id(obj: Union[str, EnsureIdentifiable]) -> str:
    return obj if isinstance(obj, str) else obj.id



from typing import Any, Generic, Optional, Type, TypeVar

from .abstract import AbstractElementModel, AbstractModelSerializer

_E = TypeVar('_E')


class DataclassSerializer(AbstractModelSerializer[_E], Generic[_E]):

    def __init__(self, t: Optional[Type[_E]] = None):
        self._model_type = self._generate_model(t) if t is not None else None

    def serialize(self, element: _E) -> AbstractElementModel[Any]:
        if self._model_type is None:
            return self.field_type(type(element)).serialize(element)
        return self._model_type.serialize(element)

    def field_type(self, element_type: Type[_E]) -> Type[AbstractElementModel[_E]]:
        if self._model_type is None:
            return self._generate_model(element_type)
        return self._model_type

    @staticmethod
    def _generate_model(t: Type[_E]) -> Type[AbstractElementModel[_E]]:
        from tdm.abstract.json_schema.model import create_model_for_type
        return create_model_for_type(t)



from typing import Collection, Dict, Mapping, Sequence, Tuple, Type, TypeVar, Union

from tdm.abstract.datamodel import AbstractDomainType, AbstractFact, AbstractMarkup, AbstractNode, AbstractNodeMention, AbstractValue, \
    BaseNodeMetadata
from tdm.helper import unfold_union, uniform_collection
from .abstract import AbstractElementSerializer
from .composite import CompositeElementSerializer
from .dataclass import DataclassSerializer
from .domain import DomainTypeSerializer
from .identifiable import IdSerializer
from .identity import IdentitySerializer
from .markup import MarkupSerializer
from .mention import NodeMentionSerializer
from .type_ import TypeSerializer
from .value import ValueSerializer
from .wrap import MappingElementSerializer, SequenceElementSerializer

_BASE_SERIALIZERS = {
    AbstractNode: IdSerializer(AbstractNode),
    AbstractFact: IdSerializer(AbstractFact),
    AbstractNodeMention: NodeMentionSerializer(),
    BaseNodeMetadata: DataclassSerializer(),
    AbstractValue: ValueSerializer(),
    AbstractMarkup: MarkupSerializer(),
    AbstractDomainType: DomainTypeSerializer()
}

_E = TypeVar('_E')
_S = TypeVar('_S')


def _prepare(serializer: AbstractElementSerializer[_E, _S], arg: Type[_E]) -> Tuple[Type[_S], _E]:
    return serializer.field_type(arg), serializer


def get_serializer(t: Type, type_vars: Dict[TypeVar, Type] = None) -> Tuple[type, AbstractElementSerializer]:
    if isinstance(t, TypeVar) and type_vars is not None and t in type_vars:
        return get_serializer(type_vars[t], type_vars)
    t = unfold_union(t)
    if len(t) > 1:
        field_types, serializers = zip(*map(lambda x: get_serializer(x, type_vars), t))
        return Union[field_types], CompositeElementSerializer.build(serializers)
    else:
        t = t[0]
    real_type, arg = uniform_collection(t)
    if real_type is type:
        return _prepare(TypeSerializer(), arg)
    if real_type is not None:
        if issubclass(real_type, Collection):
            wrapped_type, serializer = get_serializer(arg, type_vars)
            if issubclass(real_type, Mapping):
                serializer = MappingElementSerializer(real_type, serializer)
            elif issubclass(real_type, Sequence):
                serializer = SequenceElementSerializer(real_type, serializer)
            else:
                raise TypeError
            return _prepare(serializer, wrapped_type)
        raise TypeError
    possible_types = set(arg.mro()).intersection(_BASE_SERIALIZERS)
    if len(possible_types) == 1:
        return _prepare(_BASE_SERIALIZERS[possible_types.pop()], arg)
    if hasattr(arg, '__dataclass_fields__'):
        return _prepare(DataclassSerializer(), arg)
    return _prepare(IdentitySerializer(arg), arg)
